from __future__ import print_function
import datetime
import os
import unittest

import cfdm

# Build the test suite from the tests found in the test files.
testsuite_setup = unittest.TestSuite()
testsuite_setup.addTests(unittest.TestLoader().discover('.', pattern='setup_create_field*.py'))

testsuite = unittest.TestSuite()
testsuite.addTests(unittest.TestLoader().discover('.', pattern='test_*.py'))
 
# Run the test suite.
def run_test_suite_setup(verbosity=2):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    runner.run(testsuite_setup)

# Run the test suite.
def run_test_suite(verbosity=2):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    runner.run(testsuite)

if __name__ == '__main__':
    print('---------------')
    print('CFDM TEST SUITE')
    print('---------------')
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    print('Running tests from', os.path.abspath(os.curdir))
    print('')
    
    run_test_suite_setup()
    run_test_suite()



from __future__ import print_function
from builtins import (range, str)

import datetime
import tempfile
import os
import unittest
import atexit
import inspect

import numpy

import cfdm

tmpfile  = tempfile.mktemp('.cf-python_test')
tmpfiles = [tmpfile]
def _remove_tmpfiles():
    '''
'''
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass
#--- End: def
atexit.register(_remove_tmpfiles)

class read_writeTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    test_only = []
#    test_only = ['NOTHING!!!!!']
#    test_only = ['test_write_HDF_chunks']
#    test_only = ['test_read_write_unlimited']
#    test_only = ['test_read_field']
#    test_only = ['test_write_datatype']
    
    def test_read_field(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Test field keyword of cfdm.read
        filename = self.filename
        
        f = cfdm.read(filename)
        self.assertTrue(len(f) == 1, '\n'+str(f))

        f = cfdm.read(filename, create_field=['dimension_coordinate'])
        self.assertTrue(len(f) == 4, '\n'+str(f))

        f = cfdm.read(filename, create_field=['auxiliary_coordinate'])
        self.assertTrue(len(f) == 4, '\n'+str(f))
        
        f = cfdm.read(filename, create_field='cell_measure')
        self.assertTrue(len(f) == 2, '\n'+str(f))

        f = cfdm.read(filename, create_field=['field_ancillary'])
        self.assertTrue(len(f) == 4, '\n'+str(f))
                
        f = cfdm.read(filename, create_field='domain_ancillary')
        self.assertTrue(len(f) == 4, '\n'+str(f))
        
        f = cfdm.read(filename, create_field=['field_ancillary', 'auxiliary_coordinate'])
        self.assertTrue(len(f) == 7, '\n'+str(f))
        
        self.assertTrue(len(cfdm.read(filename, create_field=['domain_ancillary', 'auxiliary_coordinate'])) == 7)
        self.assertTrue(len(cfdm.read(filename, create_field=['domain_ancillary', 'cell_measure', 'auxiliary_coordinate'])) == 8)

        f = cfdm.read(filename, create_field=('field_ancillary', 'dimension_coordinate',
                                       'cell_measure', 'auxiliary_coordinate',
                                       'domain_ancillary'))
        self.assertTrue(len(f) == 14, '\n'+str(f))
    #--- End: def

    def test_read_write_format(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0]
        for fmt in ('NETCDF3_CLASSIC',
                    'NETCDF3_64BIT',
                    'NETCDF4',
                    'NETCDF4_CLASSIC'):
            cfdm.write(f, tmpfile, fmt=fmt)
            g = cfdm.read(tmpfile)[0]
            self.assertTrue(f.equals(g, traceback=True),
                            'Bad read/write of format: {}'.format(fmt))
        #--- End: for
    #--- End: def

    def test_read_write_netCDF4_compress_shuffle(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0]
        for fmt in ('NETCDF4',
                    'NETCDF4_CLASSIC'):
            for shuffle in (True, False):
                for compress in range(10):
                    cfdm.write(f, tmpfile, fmt=fmt,
                               compress=compress,
                               shuffle=shuffle)
                    g = cfdm.read(tmpfile)[0]
                    self.assertTrue(
                        f.equals(g, traceback=True),
                        'Bad read/write with lossless compression: {}, {}, {}'.format(fmt, compress, shuffle))
        #--- End: for
    #--- End: def

    def test_write_datatype(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0] 
        self.assertTrue(f.data.dtype == numpy.dtype(float))

        cfdm.write(f, tmpfile, fmt='NETCDF4', 
                 datatype={numpy.dtype(float): numpy.dtype('float32')})
        g = cfdm.read(tmpfile)[0]
        self.assertTrue(g.data.dtype == numpy.dtype('float32'), 
                        'datatype read in is '+str(g.data.dtype))
    #--- End: def

#    def test_write_HDF_chunks(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
#            
#        for fmt in ('NETCDF3_CLASSIC', 'NETCDF4'):
#            f = cfdm.read(self.filename)[0]
#            f.HDF_chunks({'T': 10000, 1: 3, 'grid_lat': 222, 45:45})
#            cfdm.write(f, tmpfile, fmt=fmt, HDF_chunksizes={'X': 6})
#        #--- End: for
#    #--- End: def

#    def test_read_write_unlimited(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
#        f = cfdm.read(self.filename)[0]
#
#        fmt = 'NETCDF4'
#        for axis in ('atmosphere_hybrid_height_coordinate', 'X', 'Y'):
#            org = f.unlimited({axis: True})
#            cfdm.write(f, tmpfile, fmt=fmt)
#            f.unlimited(org)
#            
#            g = cfdm.read(tmpfile)[0]
#            self.assertTrue(g.unlimited()[g.axis(axis, key=True)] is True,
#                            'Failed with axis={}, fmt={}'.format(axis, fmt))
#
#        fmt = 'NETCDF3_CLASSIC'
#        for axis in ('atmosphere_hybrid_height_coordinate',):
#            org = f.unlimited({axis: True})
#            cfdm.write(f, tmpfile, fmt=fmt)
#            f.unlimited(org)
#            
#            g = cfdm.read(tmpfile)[0]
#            self.assertTrue(g.unlimited()[g.axis(axis, key=True)] is True,
#                            'Failed with axis={}, fmt={}'.format(axis, fmt))
#
#        fmt = 'NETCDF4'
#        org = f.unlimited({'Y': True, 'X': True})
#        cfdm.write(f, tmpfile, fmt=fmt)
#        f.unlimited(org)
#
#        g = cfdm.read(tmpfile)[0]
#        self.assertTrue(g.unlimited()[g.axis('X', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('X', fmt))
#        self.assertTrue(g.unlimited()[g.axis('Y', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('Y', fmt))
#
#
#        fmt = 'NETCDF4'
#        org = f.unlimited({'X': False})
#        cfdm.write(f, tmpfile, fmt=fmt, unlimited=['X'])
#        f.unlimited(org)
#
#        g = cfdm.read(tmpfile)[0]
#        self.assertTrue(not g.unlimited()[g.axis('X', key=True)],
#                        'Failed with axis={}, fmt={}'.format('X', fmt))
#
#
#        fmt = 'NETCDF4'
#        org = f.unlimited({'X': True})
#        cfdm.write(f, tmpfile, fmt=fmt, unlimited=['X'])
#        f.unlimited(org)
#
#        g = cfdm.read(tmpfile)[0]
#        self.assertTrue(g.unlimited()[g.axis('X', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('X', fmt))
#
#
#        fmt = 'NETCDF4'
#        org = f.unlimited({'Y': True})
#        cfdm.write(f, tmpfile, fmt=fmt, unlimited=['X'])
#        f.unlimited(org)
#
#        g = cfdm.read(tmpfile)[0]
#        self.assertTrue(g.unlimited()[g.axis('X', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('X', fmt))
#        self.assertTrue(g.unlimited()[g.axis('Y', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('Y', fmt))
#
# 
#
#        fmt = 'NETCDF4'
#        org = f.unlimited({('X', 'Y'): True})
#        cfdm.write(f, tmpfile, fmt=fmt)
#        f.unlimited(org)
#
#        g = cfdm.read(tmpfile)[0]
#        self.assertTrue(g.unlimited()[g.axis('X', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('X', fmt))
#        self.assertTrue(g.unlimited()[g.axis('Y', key=True)] is True,
#                        'Failed with axis={}, fmt={}'.format('Y', fmt))
#
# 
#
#        fmt = 'NETCDF4'
#        org = f.unlimited({('X', 'Y'): True})
#        f.unlimited(None)
#        cfdm.write(f, tmpfile, fmt=fmt)
#        f.unlimited(org)
#
#        g = cfdm.read(tmpfile)[0]
#        self.assertTrue(not g.unlimited()[g.axis('X', key=True)],
#                        'Failed with axis={}, fmt={}'.format('X', fmt))
#        self.assertTrue(not g.unlimited()[g.axis('Y', key=True)],
#                        'Failed with axis={}, fmt={}'.format('Y', fmt))
#    #--- End: def

#--- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    unittest.main(verbosity=2)



from builtins import (range, super)

import numpy

from . import abstract
from . import mixin

class RaggedIndexedArray(mixin.RaggedIndexed,
                         abstract.CompressedArray):
    '''An underlying indexed ragged array.

A collection of features stored using an indexed ragged array combines
all features along a single dimension (the "sample dimension") such
that the values of each feature in the collection are interleaved.

The information needed to uncompress the data is stored in a "index
variable" that specifies the feature that each element of the sample
dimension belongs to.

.. versionadded:: 1.7

    '''
    def __init__(self, compressed_array=None, shape=None, size=None,
                 ndim=None, index_variable=None):
        '''**Initialization**

:Parameters:

    compressed_array: subclass of `Array`
        The compressed array.

    shape: `tuple`
        The uncompressed array dimension sizes.

    size: `int`
        Number of elements in the uncompressed array.

    ndim: `int`
        The number of uncompressed array dimensions

    index_variable: `Index`
        The index variable required to uncompress the data,
        corresponding to a CF-netCDF index variable.

        '''
        super().__init__(compressed_array=compressed_array,
                         shape=shape, size=size, ndim=ndim,
                         index_variable=index_variable,
                         compressed_dimension=0,
                         compression_type='ragged indexed')
    #--- End: def

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

Returns an subspace of the uncompressed data an independent numpy
array.

The indices that define the subspace are relative to the uncompressed
data and must be either `Ellipsis` or a sequence that contains an
index for each dimension. In the latter case, each dimension's index
must either be a `slice` object or a sequence of two or more integers.

Indexing is similar to numpy indexing. The only difference to numpy
indexing (given the restrictions on the type of indices allowed) is:

  * When two or more dimension's indices are sequences of integers
    then these indices work independently along each dimension
    (similar to the way vector subscripts work in Fortran).

        '''
        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        
        compressed_array = self._get_compressed_Array()

        # Initialise the un-sliced uncompressed array
        uarray = numpy.ma.masked_all(self.shape, dtype=self.dtype)

        # --------------------------------------------------------
        # Compression by indexed ragged array.
        #
        # The uncompressed array has dimensions (instance
        # dimension, element dimension).
        # --------------------------------------------------------
        index_array = self.get_index_variable().get_array()
        
        for i in range(uarray.shape[0]):
            sample_dimension_indices = numpy.where(index_array == i)[0]

            u_indices = (i, #slice(i, i+1),
                         slice(0, len(sample_dimension_indices)))
            
            uarray[u_indices] = compressed_array[(sample_dimension_indices,)]
        #--- End: for

        return self.get_subspace(uarray, indices, copy=True)
    #--- End: def

#--- End: class



from . import abstract


class InteriorRing(abstract.PropertiesData):
    '''An interior ring array with properties.

For polygon geometries, an individual part may define an "interior
ring", i.e. a hole that is to be omitted from the cell extent (as
would occur, for example, for a cell describing the land area of a
region containing a lake). In this case an interior ring array is
required that records whether each polygon is to be included or
excluded from the cell, supplied by an interior ring variable in
CF-netCDF. The interior ring array spans the same domain axes as its
coordinate array, with the addition of an extra ragged dimension that
indexes the geometries for each cell.

    '''

#--- End: class



from builtins import super

import numpy

from . import abstract
from . import mixin


class RaggedContiguousArray(mixin.RaggedContiguous,
                            abstract.CompressedArray):
    '''An underlying contiguous ragged array.

A collection of features stored using a contiguous ragged array
combines all features along a single dimension (the "sample
dimension") such that each feature in the collection occupies a
contiguous block.

The information needed to uncompress the data is stored in a "count
variable" that gives the size of each block.

.. versionadded:: 1.7

    '''
    def __init__(self, compressed_array=None, shape=None, size=None,
                 ndim=None, count_variable=None):
        '''**Initialization**

:Parameters:

    compressed_array: subclass of `Array`
        The compressed data.

    shape: `tuple`
        The uncompressed array dimension sizes.

    size: `int`
        Number of elements in the uncompressed array.

    ndim: `int`
        The number of uncompressed array dimensions

    count_variable: `Count`
        The count variable required to uncompress the data,
        corresponding to a CF-netCDF count variable.

        '''
        super().__init__(compressed_array=compressed_array,
                         shape=shape, size=size, ndim=ndim,
                         count_variable=count_variable,
                         compression_type='ragged contiguous',
                         compressed_dimension=0)
    #--- End: def

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

Returns an subspace of the uncompressed data an independent numpy
array.

The indices that define the subspace are relative to the uncompressed
data and must be either `Ellipsis` or a sequence that contains an
index for each dimension. In the latter case, each dimension's index
must either be a `slice` object or a sequence of two or more integers.

Indexing is similar to numpy indexing. The only difference to numpy
indexing (given the restrictions on the type of indices allowed) is:

  * When two or more dimension's indices are sequences of integers
    then these indices work independently along each dimension
    (similar to the way vector subscripts work in Fortran).

        '''
        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        
        compressed_array = self._get_compressed_Array()

        # Initialise the un-sliced uncompressed array
        uarray = numpy.ma.masked_all(self.shape, dtype=self.dtype)

        # --------------------------------------------------------
        # Compression by contiguous ragged array
        #
        # The uncompressed array has dimensions (instance
        # dimension, element dimension).
        # --------------------------------------------------------

        count_array = self.get_count_variable().get_array()
        
        start = 0 
        for i, n in enumerate(count_array):
            n = int(n)
            sample_indices = slice(start, start + n)
            
            u_indices = (i,
                         slice(0, sample_indices.stop - sample_indices.start))
            
            uarray[u_indices] = compressed_array[(sample_indices,)]
            
            start += n
        #--- End: for

        return self.get_subspace(uarray, indices, copy=True)
    #--- End: def

#--- End: class



from builtins import super

from . import mixin
from . import core


class Bounds(mixin.NetCDFVariable,
             mixin.PropertiesData,
             core.Bounds):
    '''A cell bounds component of a coordinate or domain ancillary
construct of the CF data model.

An array of cell bounds spans the same domain axes as its coordinate
array, with the addition of an extra dimension whose size is that of
the number of vertices of each cell. This extra dimension does not
correspond to a domain axis construct since it does not relate to an
independent axis of the domain. Note that, for climatological time
axes, the bounds are interpreted in a special way indicated by the
cell method constructs.

In the CF data model, cell bounds do not have their own properties
because they can not logically be different to those of the coordinate
construct itself. However, it is sometimes desired to store properties
on a CF-netCDF bounds variable, so this class supports this
capability.

.. versionadded:: 1.7

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

          *Example:*
             ``properties={'standard_name': 'grid_latitude'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    data: `Data`, optional
        Set the data. Ignored if the *source* parameter is set.
        
        The data also may be set after initialisation with the
        `set_data` method.
        
    source: optional
        Initialize the properties and data from those of *source*.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, field=None, key=None, _title=None,
             _create_title=True, _prefix=None, _level=0,
             _omit_properties=None, _axes=None, _axis_names=None):
        '''A full description of the bounds component.

Returns a description of all properties and provides selected values
of all data arrays.

.. versionadded:: 1.7

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    out: `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Bounds: ' + self.name(default='')

        return super().dump(display=display, field=field, key=key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title, _axes=_axes,
                            _axis_names=_axis_names)
    #--- End: def
    
    def name(self, default=None, ncvar=True, custom=None,
             all_names=False):
        '''TODO

        '''
        inherited_properties = self._get_component('inherited_properties', None)
        if inherited_properties:
            properties = inherited_properties.copy()
            properties.update(self.properties())

            bounds = self.copy()
            bounds.properties(properties)
            self = bounds
            
        return super().name(default=default, ncvar=ncvar,
                            custom=custom,
                            all_names=all_names)
    #--- End: def

#--- End: class



from builtins import super

from . import abstract
from . import mixin

from . import Constructs


class Domain(mixin.ConstructAccess, abstract.Container):
    '''A domain of the CF data model.

A domain represents a set of discrete "locations" in what generally
would be a multi-dimensional space, either in the real world or in a
model's simulated world.

The domain is defined collectively by other constructs included in a
field construct which describe measurement locations and cell
properties:

====================  ================================================
Construct             Description
====================  ================================================
Domain axis           Independent axes of the domain stored in
                      `DomainAxis` objects

Dimension coordinate  Cell locations stored in `DimensionCoordinate`
                      objects

Auxiliary coordinate  Cell locations stored in `AuxiliaryCoordinate`
                      objects

Coordinate reference  Coordinate systems stored in
                      `CoordinateReference` objects

Domain ancillary      Ancillary values for cell locations in
                      alternative coordinate systems stored in
                      `DomainAncillary` objects

Cell measure          Cell sizes stored in `CellMeasure` objects
====================  ================================================

    '''

    # Define the base of the identity keys for each construct type
    _construct_key_base = {'auxiliary_coordinate': 'auxiliarycoordinate',
                           'cell_measure'        : 'cellmeasure',
                           'coordinate_reference': 'coordinatereference',
                           'dimension_coordinate': 'dimensioncoordinate',
                           'domain_ancillary'    : 'domainancillary',
                           'domain_axis'         : 'domainaxis',
    }
    

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance
    #--- End: def
    
    def __init__(self, source=None, copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    source: optional
        TODO
        Initialise the domain from the constructs of *source*,
        ignoring any cell method or field ancillary constructs.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__()
        
        if source is not None:
            try:                
                constructs = source._get_constructs()
            except AttributeError:
                constructs = self._Constructs(**self._construct_key_base)
                copy = False
                _use_data = True            
            else:
                constructs = constructs.view(ignore=('cell_method',
                                                     'field_ancillary'))
        else:
            constructs = self._Constructs(**self._construct_key_base)
            copy = False
            _use_data = True

        if copy or not _use_data:
            constructs = constructs.copy(data=_use_data)
            
        self._set_component('constructs', constructs, copy=False)
    #--- End: def
    
#    def __repr__(self):
#        '''x.__repr__() <==> repr(x)
#
#        '''
#        shape = sorted([domain_axis.size for domain_axis in list(self.domain_axes().values())]#)
#
#        return '<{0}: {1}>'.format(self.__class__.__name__, shape)
#    #--- End: def
#
#    def __str__(self):
#        '''x.__str__() <==> str(x)
#
#        '''
#        return 'STRINASDAs da sa'
#    #-- End: def

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _get_constructs(self, *default):
        '''TODO

.. versionadded:: 1.7
        
        '''
        return self._get_component('constructs', *default)
    #--- End: def
    
    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, data=True):
        '''Return a deep copy.

``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

.. versionadded:: 1.7

.. seealso:: `fromconstructs`

:Parameters:

    data: `bool`, optional
        If False then do not copy data. By default data are copied.

:Returns:	

    out:
        The deep copy.

**Examples:**

>>> e = d.copy()

        '''
        return type(self)(source=self, copy=True, _use_data=data)
    #--- End: def

    def del_construct(self, cid):
        '''
        '''
        if cid in self.domain_axes():
            for k, v in self.array_constructs().items():
                if cid in self.construct_axes().get(k, ()):
                    raise ValueError("asda ;wo3in dp08hi n")
        else:
            # Remove pointers to removed construct in coordinate
            # reference constructs
            for ref in self.coordinate_references().values():
                coordinate_conversion = ref.coordinate_conversion
                for term, value in coordinate_conversion.domain_ancillaries().items():
                    if cid == value:
                        coordinate_conversion.set_domain_ancillary(term, None)
                    
                for coord_key in ref.coordinates():
                    if cid == coord_key:
                        ref.del_coordinate(coord_key)
                        break
        #--- End: if
        
        return self._get_constructs().del_construct(cid)
    #--- End: def

    @classmethod
    def fromconstructs(cls, constructs):
        '''TODO

.. versionadded:: 1.7

.. seealso:: `copy`

:Parameters:

    constructs: `Constructs`
        
:Returns:

    out: `Domain`
        TODO

**Examples:**

>>> d = Domain.fromconstructs(f._get_constructs())
        
        '''
        domain = cls()
        domain._set_component('constructs',
                              constructs.view(ignore=('cell_method',
                                                      'field_ancillary')),
                              copy=False)
        return domain
    #--- End: def
            
#--- End: class



from .container                   import Container
from .properties                  import Properties
from .propertiesdata              import PropertiesData
from .propertiesdatabounds        import PropertiesDataBounds
from .coordinate                  import Coordinate

from .parameters                  import Parameters
from .parametersdomainancillaries import ParametersDomainAncillaries



    



from builtins import object

import re


class ConstructAccess(object):
    '''Mixin class for manipulating constructs stored in a `Constructs`
object.

    '''

    def _unique_construct_names(self):
        '''TODO 

        '''    
        key_to_name = {}
        name_to_keys = {}


        for d in self._get_constructs()._constructs.values():
            name_to_keys = {}
        
            for key, construct in d.items():
                name = construct.name(default='cid%'+key)
                name_to_keys.setdefault(name, []).append(key)
                key_to_name[key] = name
    
            for name, keys in name_to_keys.items():
                if len(keys) <= 1:
                    continue
                
                for key in keys:
                    key_to_name[key] = '{0}{{{1}}}'.format(
                        name,
                        re.findall('\d+$', key)[0])
        #--- End: for
        
        return key_to_name
    #--- End: def
    
    def _unique_domain_axis_names(self):
        '''TODO 
        '''
        key_to_name = {}
        name_to_keys = {}

        for key, value in self.domain_axes().items():
            name_size = (self.domain_axis_name(key), value.get_size(''))
            name_to_keys.setdefault(name_size, []).append(key)
            key_to_name[key] = name_size

        for (name, size), keys in name_to_keys.items():
            if len(keys) == 1:
                key_to_name[keys[0]] = '{0}({1})'.format(name, size)
            else:
                for key in keys:                    
                    key_to_name[key] = '{0}{{{1}}}({2})'.format(name,
                                                                re.findall('\d+$', key)[0],
                                                                size)
        #--- End: for
        
        return key_to_name
    #--- End: def
    
    def coordinate_references(self, copy=False):
        '''Return coordinate reference constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.


**Examples:**

>>> f.coordinate_references()
{}

>>> f.coordinate_references()
{'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>}

        '''
        return self._get_constructs().constructs(
            construct_type='coordinate_reference', copy=copy)
    #--- End: def

    def coordinates(self, copy=False):
        '''Return dimension and auxiliary coordinate constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**

>>> f.coordinates()
{}

>>> f.coordinates()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >}

        '''
        out = self.dimension_coordinates(copy=copy)
        out.update(self.auxiliary_coordinates(copy=copy))
        return out
    #--- End: def

    def domain_ancillaries(self, copy=False):
        '''Return domain ancillary constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**

>>> f.domain_ancillaries()
{}

>>> f.domain_ancillaries()
{'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>}

        '''
        return self._get_constructs().constructs(
            construct_type='domain_ancillary', copy=copy)
    #--- End: def
    
    def cell_measures(self, copy=False):
        '''Return cell measure constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**

>>> f.cell_measures()
{}

>>> f.cell_measures()
{'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>}

        '''
        return self._get_constructs().constructs(
            construct_type='cell_measure', copy=copy)
    #--- End: def

    def del_construct(self, description=None, cid=None, axes=None,
                      construct_type=None):
        '''Remove a metadata construct.

The construct is identified via optional parameters. The *unique*
construct that satisfies *all* of the given criteria is removed. An
error is raised if multiple constructs satisfy all of the given
criteria.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`, `has_construct`,
             `set_construct`

:Parameters:

    description: `str`, optional
        Select constructs that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

            *Example:*
              ``description='air_pressure'`` will select constructs
              that have a "standard_name" property with the value
              "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

            *Example:*
              ``description='positive:up'`` will select constructs
              that have a "positive" property with the value "up".

            *Example:*
              ``description='foo:bar'`` will select constructs that
              have a "foo" property with the value "bar".

            *Example:*
              ``description='standard_name:air_pressure'`` will select
              constructs that have a "standard_name" property with the
              value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

            *Example:*
              ``description='measure%area'`` will select "area" cell
              measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

            *Example:* 
              ``description='cid%cellmethod1'`` will select cell
              method construct with construct identifier
              "cellmethod1". This is equivalent to
              ``cid='cellmethod1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

            *Example:*
              ``description='ncvar%lat'`` will select constructs with
              netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

            *Example:*
              ``description='ncdim%time'`` will select domain axis
              constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary1". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: `str`, optional
        Select constructs of the given type. Valid types are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Example:*
          ``axes=['domainaxis2']``

        *Example:*
          ``axes=['domainaxis0', 'domainaxis1']``

:Returns:

    out:
        The removed metadata construct.

**Examples:**

>>> c = f.del_construct('grid_latitude')
>>> c = f.del_construct('long_name:Air Pressure')
>>> c = f.del_construct('ncvar%lat)
>>> c = f.del_construct('cid%cellmeasure0')
>>> c = f.del_construct(cid='domainaxis2')
>>> c = f.del_construct(construct_type='auxiliary_coordinate',
...                     axes=['domainaxis1'])

        '''
        key = self.get_construct_id(description=description, cid=cid,
                                    construct_type=construct_type,
                                    axes=axes)
        if key is None:
            raise ValueError("as 345 345 sdjh aiua oi ")

        return self._get_constructs().del_construct(cid=key)
    #--- End: def

    def get_construct(self, description=None, cid=None, axes=None,
                      construct_type=None, copy=False):
        '''Return a metadata construct.

The construct is selected via optional parameters. The *unique*
construct that satisfies *all* of the given criteria is returned. An
error is raised if multiple constructs satisfy all of the given
criteria.

.. versionadded:: 1.7

.. seealso:: `constructs`, `del_construct`, `has_construct`,
             `set_construct`

:Parameters:

    description: `str`, optional
        Select the construct that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

          *Example:*
            ``description='air_pressure'`` will select constructs that
            have a "standard_name" property with the value
            "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

          *Example:*
            ``description='positive:up'`` will select constructs that
            have a "positive" property with the value "up".

          *Example:*
            ``description='foo:bar'`` will select constructs that have
            a "foo" property with the value "bar".

          *Example:*
            ``description='standard_name:air_pressure'`` will select
            constructs that have a "standard_name" property with the
            value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

          *Example:*
            ``description='measure%area'`` will select "area" cell
            measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

          *Example:* 
            ``description='cid%cellmethod1'`` will select cell method
            construct with construct identifier "cellmethod1". This is
            equivalent to ``cid='cellmethod1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

          *Example:*
            ``description='ncvar%lat'`` will select constructs with
            netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

          *Example:*
            ``description='ncdim%time'`` will select domain axis
            constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary0". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: `str`, optional
        Select constructs of the given type. Valid types are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Example:*
          ``axes=['domainaxis2']``

        *Example:*
          ``axes=['domainaxis0', 'domainaxis1']``

    copy: `bool`, optional
        If True then return a copy of the unique selected
        construct. By default the construct is not copied.

:Returns:

    out:
        The unique selected construct.

**Examples:**

>>> c = f.get_construct('grid_latitude')
>>> c = f.get_construct('long_name:Air Pressure')
>>> c = f.get_construct('ncvar%lat)
>>> c = f.get_construct('cid%cellmeasure0')
>>> c = f.get_construct(cid='domainaxis2')
>>> c = f.get_construct(construct_type='auxiliary_coordinate',
...                     axes=['domainaxis1'])

        '''
        out = self.constructs(description=description, cid=cid,
                              construct_type=construct_type,
                              axes=axes, copy=copy)

        if not out:
            raise ValueError("No such construct {} {} {}".format(description, construct_type, axes))
        
        _, construct = out.popitem()
        if out:
            raise ValueError("More than one construct meets criteria")
            
        return construct
    
#        return self._get_constructs().get_construct(
#            description=description, cid=cid,
#            construct_type=construct_type, axes=axes, copy=copy)            
    #--- End: def
    
    def get_construct_axes(self, description=None, cid=None,
                           axes=None, construct_type=None):
        '''Return the domain axes spanned by a metadata construct data array.

The construct is selected via optional parameters. The *unique*
construct that satisfies *all* of the given criteria is selected. If
there is not a unique construct that satisfies all of the given
criteria, then an empty tuple is returned.

.. versionadded:: 1.7

.. seealso:: `construct_axes`, `get_construct`, `set_construct_axes`

:Parameters:

    description: `str`, optional
        Select the construct that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

          *Example:*
            ``description='air_pressure'`` will select constructs that
            have a "standard_name" property with the value
            "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

          *Example:*
            ``description='positive:up'`` will select constructs that
            have a "positive" property with the value "up".

          *Example:*
            ``description='foo:bar'`` will select constructs that have
            a "foo" property with the value "bar".

          *Example:*
            ``description='standard_name:air_pressure'`` will select
            constructs that have a "standard_name" property with the
            value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

          *Example:*
            ``description='measure%area'`` will select "area" cell
            measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

          *Example:* 
            ``description='cid%cellmethod1'`` will select cell method
            construct with construct identifier "cellmethod1". This is
            equivalent to ``cid='cellmethod1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

          *Example:*
            ``description='ncvar%lat'`` will select constructs with
            netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

          *Example:*
            ``description='ncdim%time'`` will select domain axis
            constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary0". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: `str`, optional
        Select constructs of the given type. Valid types are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Example:*
          ``axes=['domainaxis2']``

        *Example:*
          ``axes=['domainaxis0', 'domainaxis1']``

:Returns:

    out: `tuple`
        The identifiers of the domain axis constructs spanned by the
        unique selected construct's data array. If there are no such
        domain axis constructs, then an empty tuple is returned

**Examples:**

TODO

        '''
        key = self.get_construct_id(description=description, cid=cid,
                                    construct_type=construct_type,
                                    axes=axes)

        if key is None:
            return ()
            
        return self.construct_axes()[key]
    #--- End: def
        
    def get_construct_id(self, description=None, cid=None, axes=None,
                         construct_type=None):
        '''Return the identifier for a metadata construct.

The construct is selected via optional parameters. The *unique*
construct that satisfies *all* of the given criteria is selected. If
there is not a unique construct that satisfies all of the given
criteria, then the `None` is returned.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`, `get_construct_axes`

:Parameters:

    description: `str`, optional
        Select the construct that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

          *Example:*
            ``description='air_pressure'`` will select constructs that
            have a "standard_name" property with the value
            "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

          *Example:*
            ``description='positive:up'`` will select constructs that
            have a "positive" property with the value "up".

          *Example:*
            ``description='foo:bar'`` will select constructs that have
            a "foo" property with the value "bar".

          *Example:*
            ``description='standard_name:air_pressure'`` will select
            constructs that have a "standard_name" property with the
            value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

          *Example:*
            ``description='measure%area'`` will select "area" cell
            measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

          *Example:* 
            ``description='cid%cellmethod1'`` will select cell method
            construct with construct identifier "cellmethod1". This is
            equivalent to ``cid='cellmethod1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

          *Example:*
            ``description='ncvar%lat'`` will select constructs with
            netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

          *Example:*
            ``description='ncdim%time'`` will select domain axis
            constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary0". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: `str`, optional
        Select constructs of the given type. Valid types are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Example:*
          ``axes=['domainaxis2']``

        *Example:*
          ``axes=['domainaxis0', 'domainaxis1']``

:Returns:

    out: `str` or `None`
        The identifier of the unique selected construct. If there are
        no such domain axis constructs, then `None` is returned.

**Examples:**

TODO

        '''
        c = self.constructs(description=description, cid=cid,
                            construct_type=construct_type, axes=axes,
                            copy=False)
        if len(c) != 1:
            # 0 or 2 or more constructs selected
            return
            
        cid, _ = c.popitem()

        # A unique construct selected
        return cid
    #--- End: def
    
    def has_construct(self, description=None, cid=None, axes=None,
                      construct_type=None):
        '''Whether a metadata construct has been set.

The construct is identified via optional parameters. If, and only if,
there is a *unique* construct that satisfies *all* of the given
criteria, then `True` is returned.

.. versionadded:: 1.7

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct`

:Parameters:

    description: `str`, optional
        Select constructs that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

            *Example:*
              ``description='air_pressure'`` will select constructs
              that have a "standard_name" property with the value
              "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

            *Example:*
              ``description='positive:up'`` will select constructs
              that have a "positive" property with the value "up".

            *Example:*
              ``description='foo:bar'`` will select constructs that
              have a "foo" property with the value "bar".

            *Example:*
              ``description='standard_name:air_pressure'`` will select
              constructs that have a "standard_name" property with the
              value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

            *Example:*
              ``description='measure%area'`` will select "area" cell
              measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

            *Example:* 
              ``description='cid%cellmethod1'`` will select cell
              method construct with construct identifier
              "cellmethod1". This is equivalent to
              ``cid='cellmethod1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

            *Example:*
              ``description='ncvar%lat'`` will select constructs with
              netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

            *Example:*
              ``description='ncdim%time'`` will select domain axis
              constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary1". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: `str`, optional
        Select constructs of the given type. Valid types are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Example:*
          ``axes=['domainaxis2']``

        *Example:*
          ``axes=['domainaxis0', 'domainaxis1']``

:Returns:

    out: `bool`
        True if a unique construct has been identified, otherwise
        False.

**Examples:**

>>> x = f.has_construct('grid_latitude')
>>> x = f.has_construct('long_name:Air Pressure')
>>> x = f.has_construct('ncvar%lat)
>>> x = f.has_construct('cid%cellmeasure0')
>>> x = f.has_construct(cid='domainaxis2')
>>> x = f.has_construct(construct_type='auxiliary_coordinate',
...                     axes=['domainaxis1'])

        '''
        out = self.constructs(description=description, cid=cid,
                              construct_type=construct_type,
                              axes=axes, copy=False)
        return len(out) == 1
    #--- End: def

    def constructs(self, description=None, cid=None, axes=None,
                   construct_type=None, copy=False):
        '''Return metadata constructs

By default all metadata constructs are returned, but a subset may be
selected via the optional parameters. If multiple parameters are
specified, then the constructs that satisfy *all* of the criteria are
returned.

.. versionadded:: 1.7

.. seealso:: `del_construct`, `get_construct`, `get_construct_axes`,
             `get_construct_id`, `has_construct`, `set_construct`

:Parameters:

    description: `str`, optional
        Select constructs that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

            *Example:*
              ``description='air_pressure'`` will select constructs
              that have a "standard_name" property with the value
              "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

            *Example:*
              ``description='positive:up'`` will select constructs
              that have a "positive" property with the value "up".

            *Example:*
              ``description='foo:bar'`` will select constructs that
              have a "foo" property with the value "bar".

            *Example:*
              ``description='standard_name:air_pressure'`` will select
              constructs that have

        a "standard_name" property with the
              value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

            *Example:*
              ``description='measure%area'`` will select "area" cell
              measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

            *Example:* 
              ``description='cid%cellmethod1'`` will select cell
              method construct with construct identifier
              "cellmethod1". This is equivalent to
              ``cid='cellmethod1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

            *Example:*
              ``description='ncvar%lat'`` will select constructs with
              netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

            *Example:*
              ``description='ncdim%time'`` will select domain axis
              constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary1". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: (sequence of) `str`, optional
        Select constructs of the given type, or types. Valid types
        are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Example:*
          ``axes=['domainaxis2']``

        *Example:*
          ``axes=['domainaxis0', 'domainaxis1']``

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.
        
        If cell method contructs, and no other construct types, have
        been selected with the *construct_type* parameter then the
        constructs are returned in an ordered dictionary
        (`collections.OrderedDict`). The order is determined by the
        order in which the cell method constructs were originally
        added.

**Examples:**

>>> f.constructs()
{}

>>> f.constructs()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >,
 'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>,
 'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>,
 'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>,
 'domainaxis0': <DomainAxis: 1>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>,
 'domainaxis3': <DomainAxis: 1>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs('grid_latitude')
{'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>}
>>> f.constructs('long_name:Grid latitude name')
{'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >}
>>> f.constructs('ncvar%b')
{'domainancillary1': <DomainAncillary: ncvar%b(1) >}
>>> f.constructs(construct_type='coordinate_reference')
{'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>}
>>> f.constructs(construct_type='cell_method')
OrderedDict([('cellmethod0', <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>),
             ('cellmethod1', <CellMethod: domainaxis3: maximum>)])
>>> f.constructs(construct_type=['cell_method', 'field_ancillary'])
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs(axes=['domainaxis0'])
{'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >}
>>> f.constructs(axes=['domainaxis0', 'domainaxis1'])
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >,
 'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs('longitude',
...              construct_type='auxiliary_coordinate', 
...              axes=['domainaxis1'])
{'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>}
>>> f.constructs('air_pressure')
{}

        '''
        return self._get_constructs().constructs(
            description=description, cid=cid,
            construct_type=construct_type, axes=axes, copy=copy)
    #--- End: def

    def domain_axes(self, copy=False):
        '''Return domain axis constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**


>>> f.domain_axes()
{}

>>> f.domain_axes()
{'domainaxis0': <DomainAxis: 1>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>,
 'domainaxis3': <DomainAxis: 1>}

        '''
        return self._get_constructs().constructs(
            construct_type='domain_axis', copy=copy)
    #--- End: def
        
    def auxiliary_coordinates(self, copy=False):
        '''Return auxiliary coordinate constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**

>>> f.auxiliary_coordinates()
{}

>>> f.auxiliary_coordinates()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >}

        '''
        return self._get_constructs().constructs(
            construct_type='auxiliary_coordinate', copy=copy)
    #--- End: def
 
    def dimension_coordinates(self, copy=False):
        '''Return dimension coordinate constructs.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**

>>> f.dimension_coordinates()
{}

>>> f.dimension_coordinates()
{'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >}

        '''
        return self._get_constructs().constructs(
            construct_type='dimension_coordinate', copy=copy)
    #--- End: def
    
    def domain_axis_name(self, axis):
        '''TODO 
        '''
        return self._get_constructs().domain_axis_name(axis)
    #--- End: def
    
#--- End: class
